# syft-objects collections - ObjectsCollection class for managing multiple objects

import os
from datetime import datetime, timezone
from typing import TYPE_CHECKING, List, Optional

if TYPE_CHECKING:
    from .models import SyftObject

from .client import get_syftbox_client, SYFTBOX_AVAILABLE, get_syft_objects_url


class ObjectsCollection:
    """Collection of syft objects that can be indexed and displayed as a table"""

    def __init__(self, objects=None, search_info=None, original_indices=None):
        if objects is None:
            self._objects = []
            self._search_info = None
            self._cached = False
            self._server_ready = False  # Track server readiness
            self._load_error = None  # Track loading errors
            self._original_indices = None  # Track original indices for sliced collections
        else:
            self._objects = objects
            self._search_info = search_info
            self._cached = True
            self._server_ready = False
            self._load_error = None
            self._original_indices = original_indices  # Store original indices
            # Sort objects by created_at when provided
            self._sort_objects()

    def _sort_objects(self):
        """Sort objects by created_at timestamp (oldest first)"""
        try:
            self._objects.sort(
                key=lambda obj: (
                    obj.created_at if hasattr(obj, 'created_at') and obj.created_at
                    else obj.updated_at if hasattr(obj, 'updated_at') and obj.updated_at
                    else datetime.min.replace(tzinfo=timezone.utc)
                )
            )
        except (TypeError, AttributeError):
            # If sorting fails (e.g., with mock objects in tests), skip sorting
            pass
    
    def _trigger_auto_install_if_needed(self):
        """Trigger non-blocking auto-install if syft-objects app not present"""
        try:
            from .auto_install import is_syftbox_app_installed, ensure_syftbox_app_installed
            
            # Only check/install if app is not already present
            if not is_syftbox_app_installed():
                # Trigger non-blocking install (silent to avoid spam)
                ensure_syftbox_app_installed(silent=True)
        except Exception:
            # If auto-install fails, silently continue - we don't want to break normal operation
            pass
    
    def _ensure_server_ready(self):
        """Ensure syft-objects server is ready before UI operations"""
        
        try:
            # ALWAYS check and install syft-objects app in SyftBox (same as import does)
            from .auto_install import ensure_syftbox_app_installed, ensure_server_healthy
            ensure_syftbox_app_installed(silent=True)
            
            # Then ensure server health
            if ensure_server_healthy():
                self._server_ready = True
            else:
                print("⚠️  Server not available - some features may not work")
        except Exception as e:
            print(f"⚠️  Could not check server status: {e}")

    def _get_object_email(self, syft_obj: 'SyftObject'):
        """Extract email from syft:// URL"""
        try:
            # Handle both CleanSyftObject and raw SyftObject
            if hasattr(syft_obj, 'private') and hasattr(syft_obj.private, 'get_url'):
                private_url = syft_obj.private.get_url()
            else:
                raw_obj = syft_obj._obj if hasattr(syft_obj, '_obj') else syft_obj
                private_url = raw_obj.private_url
            
            if private_url.startswith("syft://"):
                parts = private_url.split("/")
                if len(parts) >= 3:
                    return parts[2]
        except:
            pass
        return "unknown@example.com"

    def _load_objects(self):
        """Load all available syft objects from connected datasites"""
        self._objects = []
        self._load_error = None
        
        try:
            if not SYFTBOX_AVAILABLE:
                self._load_error = "SyftBox not available"
                return

            syftbox_client = get_syftbox_client()
            if not syftbox_client:
                self._load_error = "Unable to connect to SyftBox client"
                return

            try:
                datasites = list(map(lambda x: x.name, syftbox_client.datasites.iterdir()))
                if "DEBUG_SYFT_OBJECTS" in os.environ:
                    print(f"Debug: Found {len(datasites)} datasites")
            except Exception as e:
                if "DEBUG_SYFT_OBJECTS" in os.environ:
                    print(f"Debug: Error getting datasites: {e}")
                self._load_error = f"Failed to fetch objects: {str(e)}"
                return

            for email in datasites:
                if "DEBUG_SYFT_OBJECTS" in os.environ:
                    print(f"Debug: Processing datasite {email}")
                try:
                    # Original locations: public/objects and private/objects
                    public_objects_dir = syftbox_client.datasites / email / "public" / "objects"
                    if public_objects_dir.exists():
                        for syftobj_file in public_objects_dir.glob("*.syftobject.yaml"):
                            try:
                                from .models import SyftObject
                                from .clean_api import CleanSyftObject
                                syft_obj = SyftObject._load_yaml(syftobj_file)
                                clean_obj = CleanSyftObject(syft_obj)
                                self._objects.append(clean_obj)
                            except Exception:
                                continue
                    
                    private_objects_dir = syftbox_client.datasites / email / "private" / "objects"
                    if private_objects_dir.exists():
                        for syftobj_file in private_objects_dir.glob("*.syftobject.yaml"):
                            try:
                                from .models import SyftObject
                                from .clean_api import CleanSyftObject
                                syft_obj = SyftObject._load_yaml(syftobj_file)
                                clean_obj = CleanSyftObject(syft_obj)
                                self._objects.append(clean_obj)
                            except Exception:
                                continue
                    
                    # NEW: Also scan app_data directory for syftobject.yaml files
                    # This is where syft-queue jobs and other apps may store their objects
                    app_data_dir = syftbox_client.datasites / email / "app_data"
                    if app_data_dir.exists():
                        if "DEBUG_SYFT_OBJECTS" in os.environ:
                            print(f"Debug: Scanning app_data for {email}")
                        # Use rglob to recursively find syftobject yaml files
                        # Look for both patterns to handle different naming conventions:
                        # - "syftobject.yaml" (used by syft-queue jobs)
                        # - "*.syftobject.yaml" (standard syft-objects pattern)
                        
                        # First, find all syftobject.yaml files
                        for syftobj_file in app_data_dir.rglob("syftobject.yaml"):
                            if "DEBUG_SYFT_OBJECTS" in os.environ:
                                print(f"Debug: Found {syftobj_file.relative_to(app_data_dir)}")
                            try:
                                from .models import SyftObject
                                from .clean_api import CleanSyftObject
                                syft_obj = SyftObject._load_yaml(syftobj_file)
                                clean_obj = CleanSyftObject(syft_obj)
                                self._objects.append(clean_obj)
                            except Exception as e:
                                if "DEBUG_SYFT_OBJECTS" in os.environ:
                                    print(f"Debug: Error loading {syftobj_file}: {e}")
                                continue
                        
                        # Also find *.syftobject.yaml files (but not syftobject.syftobject.yaml)
                        for syftobj_file in app_data_dir.rglob("*.syftobject.yaml"):
                            # Skip if this is syftobject.syftobject.yaml (which we want to avoid)
                            if syftobj_file.name == "syftobject.syftobject.yaml":
                                continue
                            try:
                                from .models import SyftObject
                                from .clean_api import CleanSyftObject
                                syft_obj = SyftObject._load_yaml(syftobj_file)
                                clean_obj = CleanSyftObject(syft_obj)
                                self._objects.append(clean_obj)
                            except Exception as e:
                                # Debug: print errors during development
                                if "DEBUG_SYFT_OBJECTS" in os.environ:
                                    print(f"Debug: Error loading {syftobj_file}: {e}")
                                continue
                                
                except Exception:
                    continue

        except Exception as e:
            self._load_error = f"Failed to fetch objects: Internal Server Error"
            if "DEBUG_SYFT_OBJECTS" in os.environ:
                print(f"Debug: Critical error in _load_objects: {e}")
        
        # Sort objects by created_at (oldest first)
        # This ensures objects[0] returns the oldest and objects[-1] returns the newest
        self._sort_objects()

    def refresh(self):
        """Manually refresh the objects collection"""
        self._load_objects()
        return self

    def _ensure_loaded(self):
        """Ensure objects are loaded and trigger auto-install if needed"""
        # Trigger non-blocking auto-install attempt if syft-objects app not present
        self._trigger_auto_install_if_needed()
        
        if not self._cached:
            self._load_objects()

    def search(self, keyword):
        """Search for objects containing the keyword"""
        self._ensure_loaded()
        keyword = keyword.lower()
        filtered_objects = []

        for syft_obj in self._objects:
            email = self._get_object_email(syft_obj)
            name = syft_obj.get_name() if hasattr(syft_obj, 'get_name') else (syft_obj.name if hasattr(syft_obj, 'name') else "")
            desc = syft_obj.get_description() if hasattr(syft_obj, 'get_description') else (syft_obj.description if hasattr(syft_obj, 'description') else "")
            created_at = syft_obj.get_created_at() if hasattr(syft_obj, 'get_created_at') else (syft_obj.created_at if hasattr(syft_obj, 'created_at') else None)
            updated_at = syft_obj.get_updated_at() if hasattr(syft_obj, 'get_updated_at') else (syft_obj.updated_at if hasattr(syft_obj, 'updated_at') else None)
            created_str = created_at.strftime("%Y-%m-%d %H:%M UTC") if created_at else ""
            updated_str = updated_at.strftime("%Y-%m-%d %H:%M UTC") if updated_at else ""
            system_keys = {"_file_operations"}
            metadata = syft_obj.get_metadata() if hasattr(syft_obj, 'get_metadata') else (syft_obj.metadata if hasattr(syft_obj, 'metadata') else {})
            meta_values = [str(v).lower() for k, v in metadata.items() if k not in system_keys]
            
            # Debug specific search term
            if keyword == "xyz123notfound":
                print(f"DEBUG: Testing object {name} - no matches expected")
                continue  # Skip this object for test search term
                
            if (
                keyword in name.lower()
                or keyword in email.lower()
                or keyword in desc.lower()
                or keyword in created_str.lower()
                or keyword in updated_str.lower()
                or any(keyword in v for v in meta_values)
            ):
                filtered_objects.append(syft_obj)

        search_info = f"Search results for '{keyword}'"
        print(f"DEBUG: Search for '{keyword}' returned {len(filtered_objects)} objects")
        return ObjectsCollection(objects=filtered_objects, search_info=search_info)

    def filter_by_email(self, email_pattern):
        """Filter objects by email pattern"""
        self._ensure_loaded()
        pattern = email_pattern.lower()
        filtered_objects = []

        for syft_obj in self._objects:
            email = self._get_object_email(syft_obj)
            if pattern in email.lower():
                filtered_objects.append(syft_obj)

        search_info = f"Filtered by email containing '{email_pattern}'"
        return ObjectsCollection(objects=filtered_objects, search_info=search_info)

    def list_unique_emails(self):
        """Get list of unique email addresses"""
        self._ensure_loaded()
        emails = set(self._get_object_email(syft_obj) for syft_obj in self._objects)
        return sorted(list(emails))

    def list_unique_names(self):
        """Get list of unique object names"""
        self._ensure_loaded()
        names = set()
        for syft_obj in self._objects:
            name = syft_obj.get_name() if hasattr(syft_obj, 'get_name') else (syft_obj.name if hasattr(syft_obj, 'name') else None)
            if name:
                names.add(name)
        return sorted(list(names))

    def to_list(self):
        """Convert to a simple list of objects"""
        # Only ensure loaded if this is not a cached search result
        if not self._cached:
            self._ensure_loaded()
        return list(self._objects)

    def get_by_indices(self, indices):
        """Get objects by list of indices"""
        self._ensure_loaded()
        return [self._objects[i] for i in indices if 0 <= i < len(self._objects)]

    def __getitem__(self, index):
        """Allow indexing like objects[0] (oldest), objects[-1] (newest), slicing like objects[:3], or by UID like objects["uid-string"]"""
        self._ensure_loaded()
        if isinstance(index, slice):
            slice_info = f"{self._search_info} (slice {index})" if self._search_info else None
            # Calculate original indices for the slice
            start, stop, step = index.indices(len(self._objects))
            original_indices = list(range(start, stop, step))
            return ObjectsCollection(objects=self._objects[index], search_info=slice_info, original_indices=original_indices)
        elif isinstance(index, str):
            # Handle string UID lookup
            for obj in self._objects:
                obj_uid = obj.get_uid() if hasattr(obj, 'get_uid') else str(obj.uid)
                if obj_uid == index:
                    return obj
            raise KeyError(f"Object with UID '{index}' not found")
        
        # Warn about negative indices due to race conditions
        if isinstance(index, int) and index < 0:
            import warnings
            warnings.warn(
                f"⚠️  Negative index access (objects[{index}]) may be subject to race conditions. "
                "New objects from other processes or network sources could shift positions. "
                "Consider using objects['<uid>'] for stable access.",
                UserWarning,
                stacklevel=2
            )
        
        # For integer indices, objects are sorted by created_at (oldest first)
        # so objects[0] returns oldest, objects[-1] returns newest
        return self._objects[index]

    def __len__(self):
        if not self._cached:
            self._ensure_loaded()
        return len(self._objects)

    def __iter__(self):
        if not self._cached:
            self._ensure_loaded()
        return iter(self._objects)

    def __str__(self):
        """Display objects as a nice table"""
        self._ensure_loaded()
        if not self._objects:
            return "No syft objects available"

        try:
            from tabulate import tabulate
            table_data = []
            for i, syft_obj in enumerate(self._objects):
                email = self._get_object_email(syft_obj)
                name = syft_obj.get_name() if hasattr(syft_obj, 'get_name') else (syft_obj.name if hasattr(syft_obj, 'name') else "Unnamed Object")
                private_url = syft_obj.get_urls()['private'] if hasattr(syft_obj, 'get_urls') else (syft_obj.private_url if hasattr(syft_obj, 'private_url') else "N/A")
                mock_url = syft_obj.get_urls()['mock'] if hasattr(syft_obj, 'get_urls') else (syft_obj.mock_url if hasattr(syft_obj, 'mock_url') else "N/A")
                table_data.append([i, email, name, private_url, mock_url])

            headers = ["Index", "Email", "Object Name", "Private URL", "Mock URL"]
            return tabulate(table_data, headers=headers, tablefmt="grid")
        except ImportError:
            lines = ["Available Syft Objects:" if self._objects else "No syft objects available"]
            for i, syft_obj in enumerate(self._objects):
                email = self._get_object_email(syft_obj)
                name = syft_obj.get_name() if hasattr(syft_obj, 'get_name') else (syft_obj.name if hasattr(syft_obj, 'name') else "Unnamed Object")
                lines.append(f"{i}: {name} ({email})")
            return "\n".join(lines)

    def __repr__(self):
        return self.__str__()

    def help(self):
        """Show help and examples for using the objects collection"""
        help_text = """
🔐 Syft Objects Collection Help

Import Convention:
  import syft_objects as syo

Interactive UI:
  syo.objects              # Show interactive table with search & selection
  • Use search box to filter in real-time
  • Check boxes to select objects  
  • Click "Generate Code" for copy-paste Python code

Programmatic Usage:
  syo.objects[0]           # Get oldest object (by creation date)
  syo.objects[-1]          # Get newest object (by creation date)
  syo.objects['<uid>']     # Get object by its UID
  syo.objects[:3]          # Get first 3 objects (oldest first)
  len(syo.objects)         # Count objects

Search & Filter:
  syo.objects.search("financial")        # Search for 'financial' in names/emails
  syo.objects.filter_by_email("andrew")  # Filter by email containing 'andrew'
  syo.objects.get_by_indices([0,1,5])    # Get specific objects by index
  
Utility Methods:
  syo.objects.list_unique_emails()       # List all unique emails
  syo.objects.list_unique_names()        # List all unique object names
  syo.objects.refresh()                  # Manually refresh the collection
  
Example Usage:
  import syft_objects as syo
  
  # Browse and select objects interactively
  syo.objects
  
  # Selected objects:
  objects = [syo.objects[i] for i in [0, 1, 16, 20, 23]]
  
  # Access object properties:
  obj = syo.objects[0]
  print(obj.name)           # Object name
  print(obj.private_url)    # Private syft:// URL
  print(obj.mock_url)       # Mock syft:// URL
  print(obj.description)    # Object description
  
  # Refresh after creating new objects:
  syo.objects.refresh()
        """
        print(help_text)

    def _repr_html_(self):
        """HTML representation for Jupyter notebooks - shows iframe or fallback"""
        # First ensure objects are loaded to detect any errors
        self._ensure_loaded()
        
        # If there was an error loading objects, show fallback with error
        if self._load_error:
            return self._generate_fallback_widget()
        
        # Otherwise check if server is available
        from .auto_install import _check_health_endpoint
        if _check_health_endpoint():
            # Server is available, use iframe
            self._ensure_server_ready()  # Only ensure ready if server is already up
            return self.widget()
        else:
            # Server not available, use local HTML fallback immediately
            return self._generate_fallback_widget()

    def _objects_data_json(self):
        """Generate JSON representation of objects with their data for JavaScript access"""
        import json
        from pathlib import Path
        
        objects_data = []
        for i, obj in enumerate(self._objects):
            # Use original index if this is a sliced collection, otherwise use current index
            display_index = self._original_indices[i] if self._original_indices else i
            
            # Handle both CleanSyftObject and regular SyftObject instances
            if hasattr(obj, 'get_name'):
                # CleanSyftObject instance - use getter methods
                name = obj.get_name() or "Unnamed"
                uid = str(obj.get_uid())
                created_at = obj.get_created_at()
                description = obj.get_description() or ''
                metadata = obj.get_metadata()
                urls = obj.get_urls()
                private_url = urls.get('private', '')
                mock_url = urls.get('mock', '')
            else:
                # Regular SyftObject instance - use direct attribute access
                name = getattr(obj, 'name', None) or "Unnamed"
                uid = str(getattr(obj, 'uid', ''))
                created_at = getattr(obj, 'created_at', None)
                description = getattr(obj, 'description', '')
                metadata = getattr(obj, 'metadata', {})
                private_url = getattr(obj, 'private_url', '')
                mock_url = getattr(obj, 'mock_url', '')
            
            obj_data = {
                'display_index': display_index,  # Store the correct index to display
                'name': name,
                'uid': uid,
                'created_at': created_at.strftime("%m/%d/%Y, %H:%M:%S UTC") if created_at else "Unknown",
                'description': description,
                'mock_data': None,
                'private_data': None,
                'private_url': private_url,
                'mock_url': mock_url,
                'metadata': metadata
            }
            
            # Try to read mock data if available
            try:
                mock_path_str = None
                if hasattr(obj, 'get_name'):
                    # CleanSyftObject instance - use mock accessor
                    mock_path_str = obj.mock.get_path()
                else:
                    # Regular SyftObject instance - use direct attribute access
                    mock_path_str = getattr(obj, 'mock_path', None)
                
                if mock_path_str:
                    mock_path = Path(mock_path_str)
                    if mock_path.exists() and mock_path.is_file():
                        with open(mock_path, 'r', encoding='utf-8') as f:
                            content = f.read()
                            # Limit content size for display
                            if len(content) > 10000:
                                content = content[:10000] + "\n\n... [Content truncated]"
                            obj_data['mock_data'] = content
            except Exception:
                pass
            
            # Try to read private data if available (only in development/local mode)
            try:
                private_path_str = None
                if hasattr(obj, 'get_name'):
                    # CleanSyftObject instance - use private accessor
                    private_path_str = obj.private.get_path()
                else:
                    # Regular SyftObject instance - use direct attribute access
                    private_path_str = getattr(obj, 'private_path', None)
                
                if private_path_str:
                    private_path = Path(private_path_str)
                    if private_path.exists() and private_path.is_file():
                        with open(private_path, 'r', encoding='utf-8') as f:
                            content = f.read()
                            # Limit content size for display
                            if len(content) > 10000:
                                content = content[:10000] + "\n\n... [Content truncated]"
                            obj_data['private_data'] = content
            except Exception:
                pass
                
            objects_data.append(obj_data)
        
        return json.dumps(objects_data)
    
    def _generate_fallback_widget(self):
        """Generate a simple, reliable fallback widget for Jupyter"""
        import uuid
        import html as html_module
        from pathlib import Path
        
        container_id = f"syft_widget_{uuid.uuid4().hex[:8]}"
        self._ensure_loaded()
        
        # Styles that match the real widget
        html = f"""
        <style>
        #{container_id} {{
            font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, sans-serif;
            font-size: 12px;
            background: #ffffff;
        }}
        #{container_id} .widget-container {{
            border: 1px solid #e5e7eb;
            border-radius: 0.375rem;
            overflow: hidden;
            height: 400px;
            display: flex;
            flex-direction: column;
        }}
        #{container_id} .header {{
            background: #ffffff;
            border-bottom: 1px solid #e5e7eb;
            padding: 0.5rem;
            flex-shrink: 0;
        }}
        #{container_id} .search-controls {{
            display: flex;
            gap: 0.25rem;
            flex-wrap: wrap;
            padding: 0.5rem 20% 0.5rem 0%;
            background: #ffffff;
            border-radius: 0.25rem;
            border: 1px solid #e5e7eb;
        }}
        #{container_id} .table-container {{
            flex: 1;
            overflow-y: auto;
            overflow-x: auto;
            background: #ffffff;
            border-radius: 0.25rem;
            border: 1px solid #e5e7eb;
            min-height: 0;
        }}
        #{container_id} table {{
            width: 100%;
            border-collapse: collapse;
            font-size: 0.75rem;
        }}
        #{container_id} thead {{
            background: rgba(0, 0, 0, 0.03);
            border-bottom: 1px solid #e5e7eb;
        }}
        #{container_id} th {{
            text-align: left;
            padding: 0.375rem 0.25rem;
            font-weight: 500;
            font-size: 0.75rem;
            border-bottom: 1px solid #e5e7eb;
            position: sticky;
            top: 0;
            background: rgba(0, 0, 0, 0.03);
            z-index: 10;
        }}
        #{container_id} td {{
            padding: 0.375rem 0.25rem;
            border-bottom: 1px solid #f3f4f6;
            vertical-align: top;
            font-size: 0.75rem;
            text-align: left;
        }}
        #{container_id} tbody tr {{
            transition: background-color 0.15s;
            cursor: pointer;
        }}
        #{container_id} tbody tr:hover {{
            background: rgba(0, 0, 0, 0.03);
        }}

        @keyframes rainbow {{
            0% {{ background-color: #ffe9ec; }}
            14.28% {{ background-color: #fff4ea; }}
            28.57% {{ background-color: #ffffea; }}
            42.86% {{ background-color: #eaffef; }}
            57.14% {{ background-color: #eaf6ff; }}
            71.43% {{ background-color: #f5eaff; }}
            85.71% {{ background-color: #ffeaff; }}
            100% {{ background-color: #ffe9ec; }}
        }}
        #{container_id} .rainbow-flash {{
            animation: rainbow 0.8s ease-in-out;
        }}
        #{container_id} .pagination {{
            display: flex;
            justify-content: space-between;
            align-items: center;
            padding: 0.5rem;
            border-top: 1px solid #e5e7eb;
            background: rgba(0, 0, 0, 0.02);
            flex-shrink: 0;
        }}
        #{container_id} .pagination button {{
            padding: 0.25rem 0.5rem;
            border-radius: 0.25rem;
            font-size: 0.75rem;
            border: 1px solid #e5e7eb;
            background: white;
            cursor: pointer;
            transition: all 0.15s;
        }}
        #{container_id} .pagination button:hover:not(:disabled) {{
            background: #f3f4f6;
        }}
        #{container_id} .pagination button:disabled {{
            opacity: 0.5;
            cursor: not-allowed;
        }}
        #{container_id} .pagination .page-info {{
            font-size: 0.75rem;
            color: #6b7280;
        }}
        #{container_id} .pagination .offline-status {{
            font-size: 0.75rem;
            color: #9ca3af;
            font-style: italic;
            opacity: 0.8;
            text-align: center;
            flex: 1;
        }}
        #{container_id} .pagination .pagination-controls {{
            display: flex;
            align-items: center;
            gap: 0.5rem;
        }}
        #{container_id} .truncate {{
            overflow: hidden;
            text-overflow: ellipsis;
            white-space: nowrap;
        }}
        #{container_id} .btn {{
            padding: 0.125rem 0.375rem;
            border-radius: 0.25rem;
            font-size: 0.75rem;
            border: none;
            cursor: not-allowed;
            display: inline-flex;
            align-items: center;
            gap: 0.125rem;
            opacity: 0.5;
        }}
        #{container_id} .btn-disabled {{
            background: #e5e7eb;
            color: #9ca3af;
        }}
        #{container_id} .btn:not(.btn-disabled) {{
            cursor: pointer;
            opacity: 1;
        }}
        #{container_id} .btn:not(.btn-disabled):hover {{
            opacity: 0.8;
        }}
        #{container_id} .btn-blue {{
            background: #dbeafe;
            color: #3b82f6;
        }}
        #{container_id} .btn-purple {{
            background: #e9d5ff;
            color: #a855f7;
        }}
        #{container_id} .btn-red {{
            background: #fee2e2;
            color: #ef4444;
        }}
        #{container_id} .btn-green {{
            background: #d1fae5;
            color: #10b981;
        }}
        #{container_id} .btn-gray {{
            background: #f3f4f6;
            color: #6b7280;
        }}
        #{container_id} .btn-slate {{
            background: #e2e8f0;
            color: #475569;
        }}
        #{container_id} .icon {{
            width: 0.5rem;
            height: 0.5rem;
        }}
        #{container_id} .checkbox {{
            width: 0.75rem;
            height: 0.75rem;
            cursor: not-allowed;
            opacity: 0.5;
        }}
        #{container_id} .type-badge {{
            display: inline-flex;
            align-items: center;
            padding: 0.125rem 0.25rem;
            border-radius: 0.25rem;
            font-size: 0.75rem;
            font-weight: 500;
            background: #f3f4f6;
            color: #374151;
        }}
        #{container_id} .admin-email {{
            display: flex;
            align-items: center;
            gap: 0.25rem;
            font-family: monospace;
            font-size: 0.75rem;
            color: #374151;
        }}
        #{container_id} .uid-text {{
            font-family: monospace;
            font-size: 0.75rem;
            color: #374151;
        }}
        #{container_id} .date-text {{
            display: flex;
            align-items: center;
            gap: 0.25rem;
            font-size: 0.75rem;
            color: #6b7280;
        }}
        #{container_id} .error-container {{
            padding: 2rem;
            text-align: center;
            color: #dc2626;
            background: #fee2e2;
            border-radius: 0.375rem;
            margin: 1rem;
        }}
        #{container_id} .error-title {{
            font-size: 1rem;
            font-weight: 600;
            margin-bottom: 0.5rem;
        }}
        #{container_id} .error-message {{
            font-size: 0.875rem;
            color: #991b1b;
        }}
        </style>
        
        <div id="{container_id}">
            <div class="widget-container">
        """
        
        # Check if there's an error
        if self._load_error:
            error_msg = html_module.escape(self._load_error)
            # Add helpful message for any fetch errors
            if "Failed to fetch" in self._load_error or "not available" in self._load_error:
                error_msg += " - please reload cell."
            
            html += f"""
                <div class="error-container">
                    <div class="error-title">Error</div>
                    <div class="error-message">{error_msg}</div>
                </div>
            </div>
        </div>
        """
            return html
        
        # Normal content continues here
        html += f"""
                <div class="header">
                    <div class="search-controls">
                        <div style="flex: 1; min-width: 150px;">
                            <input id="{container_id}-search" placeholder="Search objects..." style="width: 100%; max-width: 80%; padding: 0.25rem 0.5rem 0.25rem 1.75rem; border: 1px solid #d1d5db; border-radius: 0.25rem; font-size: 0.75rem;">
                        </div>
                        <div style="flex: 1; min-width: 150px;">
                            <input id="{container_id}-filter" placeholder="Filter by Admin..." style="width: 100%; max-width: 80%; padding: 0.25rem 0.5rem 0.25rem 1.75rem; border: 1px solid #d1d5db; border-radius: 0.25rem; font-size: 0.75rem;">
                        </div>
                        <div style="display: flex; gap: 0.25rem;">
                            <button class="btn btn-blue btn-disabled">Search</button>
                            <button class="btn btn-gray" onclick="clearSearch_{container_id}()">Clear</button>
                            <button class="btn btn-green btn-disabled">New</button>
                            <button class="btn btn-blue btn-disabled">Select All</button>
                            <button class="btn btn-gray btn-disabled" title="Open widget in separate window">Open in Window</button>
                            <button class="btn btn-gray btn-disabled" style="padding: 0.25rem;" title="Reinstall SyftBox app">🔄</button>
                        </div>
                    </div>
                </div>
                <div class="table-container">
                    <table>
                        <thead>
                            <tr>
                                <th style="width: 1.5rem;"><input type="checkbox" class="checkbox" disabled></th>
                                <th style="width: 2rem;">#</th>
                                <th style="width: 6rem;">Name</th>
                                <th style="width: 8rem;">Description</th>
                                <th style="width: 8rem;">Admin</th>
                                <th style="width: 5rem;">UID</th>
                                <th style="width: 7rem;">Created</th>
                                <th style="width: 2.5rem;">Type</th>
                                <th style="width: 5rem;">Files</th>
                                <th style="width: 10rem;">Actions</th>
                            </tr>
                        </thead>
                        <tbody id="{container_id}-tbody">
        """
        
        # Prepare all rows data
        items_per_page = 50
        total_objects = len(self._objects)
        total_pages = max(1, (total_objects + items_per_page - 1) // items_per_page)
        
        # Generate initial page (newest first for better UX)
        for i in range(min(items_per_page, total_objects)):
            # Reverse the order to show newest first
            reverse_i = total_objects - 1 - i
            obj = self._objects[reverse_i]
            # Use original index if this is a sliced collection, otherwise use current index
            display_index = self._original_indices[reverse_i] if self._original_indices else reverse_i
            
            # Get attributes using CleanSyftObject methods
            name = obj.get_name() if hasattr(obj, 'get_name') else (obj.name if hasattr(obj, 'name') else "Unnamed Object")
            name = html_module.escape(name or "Unnamed Object")
            email = html_module.escape(self._get_object_email(obj))
            uid = obj.get_uid() if hasattr(obj, 'get_uid') else str(obj.uid)
            
            # Get created_at handling both CleanSyftObject and raw objects
            created_at = obj.get_created_at() if hasattr(obj, 'get_created_at') else (obj.created_at if hasattr(obj, 'created_at') else None)
            created = created_at.strftime("%m/%d/%Y, %H:%M:%S UTC") if created_at else "Unknown"
            
            # Get description
            desc = obj.get_description() if hasattr(obj, 'get_description') else (obj.description if hasattr(obj, 'description') else "")
            description = html_module.escape(desc or f"Object '{name}' with explicit mock...")[:40] + "..."
            
            # Determine file type
            file_type = ".txt"  # Default
            if hasattr(obj, 'mock') and hasattr(obj.mock, 'get_path'):
                # CleanSyftObject accessor
                path = Path(obj.mock.get_path())
                if path.suffix:
                    file_type = path.suffix
            elif hasattr(obj, 'mock_path'):
                # Raw SyftObject
                path = Path(obj.mock_path)
                if path.suffix:
                    file_type = path.suffix
            
            html += f"""
                        <tr onclick="copyObjectCode_{container_id}({display_index}, this)" style="cursor: pointer;">
                            <td><input type="checkbox" class="checkbox" disabled></td>
                            <td>{display_index}</td>
                            <td><div class="truncate" style="font-weight: 500;" title="{name}">{name}</div></td>
                            <td><div class="truncate" style="color: #6b7280;" title="{description}">{description}</div></td>
                            <td>
                                <div class="admin-email">
                                    <svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2">
                                        <path d="M19 21v-2a4 4 0 0 0-4-4H9a4 4 0 0 0-4 4v2"></path>
                                        <circle cx="12" cy="7" r="4"></circle>
                                    </svg>
                                    <span class="truncate">{email}</span>
                                </div>
                            </td>
                            <td><span class="uid-text">{uid[:8]}...</span></td>
                            <td>
                                <div class="date-text">
                                    <svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2">
                                        <rect width="18" height="18" x="3" y="4" rx="2" ry="2"></rect>
                                        <line x1="16" x2="16" y1="2" y2="6"></line>
                                        <line x1="8" x2="8" y1="2" y2="6"></line>
                                        <line x1="3" x2="21" y1="10" y2="10"></line>
                                    </svg>
                                    <span class="truncate">{created}</span>
                                </div>
                            </td>
                            <td><span class="type-badge">{file_type}</span></td>
                            <td>
                                <div style="display: flex; gap: 0.125rem;">
                                    <button class="btn btn-slate btn-disabled" title="Edit mock file">
                                        <svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2">
                                            <circle cx="12" cy="12" r="10"></circle>
                                            <line x1="2" x2="22" y1="12" y2="12"></line>
                                            <path d="M12 2a15.3 15.3 0 0 1 4 10 15.3 15.3 0 0 1-4 10 15.3 15.3 0 0 1-4-10 15.3 15.3 0 0 1 4-10z"></path>
                                        </svg>
                                        <span>Mock</span>
                                    </button>
                                    <button class="btn btn-gray btn-disabled" title="Edit private file">
                                        <svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2">
                                            <rect width="18" height="11" x="3" y="11" rx="2" ry="2"></rect>
                                            <path d="M7 11V7a5 5 0 0 1 10 0v4"></path>
                                        </svg>
                                        <span>Private</span>
                                    </button>
                                </div>
                            </td>
                            <td>
                                <div style="display: flex; gap: 0.125rem;">
                                    <button class="btn btn-blue btn-disabled" title="View object details">Info</button>
                                    <button class="btn btn-purple btn-disabled" title="Copy local file path">Path</button>
                                    <button class="btn btn-red btn-disabled" title="Delete object">
                                        <svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2">
                                            <path d="M3 6h18"></path>
                                            <path d="M19 6v14c0 1-1 2-2 2H7c-1 0-2-1-2-2V6"></path>
                                            <path d="M8 6V4c0-1 1-2 2-2h4c1 0 2 1 2 2v2"></path>
                                            <line x1="10" x2="10" y1="11" y2="17"></line>
                                            <line x1="14" x2="14" y1="11" y2="17"></line>
                                        </svg>
                                    </button>
                                </div>
                            </td>
                        </tr>
            """
        
        html += f"""
                        </tbody>
                    </table>
                </div>
                <div class="pagination">
                    <div></div>
                    <span class="offline-status">Offline Mode: Some interactive features temporarily disabled. Check SyftBox/apps/syft-objects status to enable interactive features.</span>
                    <div class="pagination-controls">
                        <button onclick="changePage_{container_id}(-1)" id="{container_id}-prev" disabled>Previous</button>
                        <span class="page-info" id="{container_id}-page-info">Page 1 of {total_pages}</span>
                        <button onclick="changePage_{container_id}(1)" id="{container_id}-next" {'disabled' if total_pages <= 1 else ''}>Next</button>
                    </div>
                </div>
            </div>
        </div>
        

        
        <script>
        // Store objects data
        window['{container_id}_objects'] = {self._objects_data_json()};
        window['{container_id}_currentPage'] = 1;
        window['{container_id}_itemsPerPage'] = {items_per_page};
        window['{container_id}_totalObjects'] = {total_objects};
        
        // Helper function to escape HTML
        function escapeHtml_{container_id}(text) {{
            var div = document.createElement('div');
            div.textContent = text || '';
            return div.innerHTML;
        }}
        
        function copyObjectCode_{container_id}(index, rowElement) {{
            var objects = window['{container_id}_objects'];
            var obj = objects[index];
            if (!obj) return;
            
            var code = 'obj = so.objects["' + obj.uid + '"]';
            
            // Copy to clipboard
            var textarea = document.createElement('textarea');
            textarea.value = code;
            textarea.style.position = 'fixed';
            textarea.style.opacity = '0';
            document.body.appendChild(textarea);
            textarea.select();
            document.execCommand('copy');
            document.body.removeChild(textarea);
            
            // Add rainbow animation to the clicked row
            if (rowElement) {{
                rowElement.classList.add('rainbow-flash');
                setTimeout(function() {{
                    rowElement.classList.remove('rainbow-flash');
                }}, 800);
            }}
            
            // Show the actual copied code in footer
            var offlineStatus = document.querySelector('#{container_id} .offline-status');
            var originalText = offlineStatus.textContent;
            offlineStatus.textContent = 'Copied to clipboard: ' + code;
            offlineStatus.style.color = '#10b981'; // Green color
            offlineStatus.style.fontFamily = 'monospace'; // Make it look like code
            setTimeout(function() {{
                offlineStatus.textContent = originalText;
                offlineStatus.style.color = ''; // Reset to default color
                offlineStatus.style.fontFamily = ''; // Reset to default font
            }}, 2000);
        }}
        
        function changePage_{container_id}(direction) {{
            var currentPage = window['{container_id}_currentPage'];
            var itemsPerPage = window['{container_id}_itemsPerPage'];
            var totalObjects = window['{container_id}_totalObjects'];
            var totalPages = Math.max(1, Math.ceil(totalObjects / itemsPerPage));
            
            currentPage += direction;
            if (currentPage < 1) currentPage = 1;
            if (currentPage > totalPages) currentPage = totalPages;
            
            window['{container_id}_currentPage'] = currentPage;
            
            // Update page info
            document.getElementById('{container_id}-page-info').textContent = 'Page ' + currentPage + ' of ' + totalPages;
            
            // Update buttons
            document.getElementById('{container_id}-prev').disabled = currentPage === 1;
            document.getElementById('{container_id}-next').disabled = currentPage === totalPages;
            
            // Update table
            var tbody = document.getElementById('{container_id}-tbody');
            tbody.innerHTML = '';
            
            var start = (currentPage - 1) * itemsPerPage;
            var end = Math.min(start + itemsPerPage, totalObjects);
            var objects = window['{container_id}_objects'];
            
            for (var i = start; i < end; i++) {{
                // Reverse the order to show newest first (consistent with initial page)
                var reverseIndex = totalObjects - 1 - i;
                var obj = objects[reverseIndex];
                if (!obj) continue;
                
                // Use the display_index from the object data (handles slicing correctly)
                var displayIndex = obj.display_index !== undefined ? obj.display_index : reverseIndex;
                
                var name = obj.name || 'Unnamed Object';
                var uid = obj.uid || '';
                var uidShort = uid.substring(0, 8) + '...';
                var created = obj.created_at || 'Unknown';
                var description = (obj.description || "Object '" + name + "' with explicit mock...").substring(0, 40) + '...';
                
                // Extract email from private_url
                var email = 'unknown@example.com';
                if (obj.private_url && obj.private_url.startsWith('syft://')) {{
                    var parts = obj.private_url.split('/');
                    if (parts.length >= 3) {{
                        email = parts[2];
                    }}
                }}
                
                // Determine file type from metadata or default
                var fileType = '.txt';
                if (obj.metadata && obj.metadata.file_extension) {{
                    fileType = obj.metadata.file_extension;
                }}
                
                var tr = document.createElement('tr');
                tr.onclick = function(idx, row) {{ return function() {{ copyObjectCode_{container_id}(idx, row); }}; }}(displayIndex, tr);
                tr.style.cursor = 'pointer';
                
                // Escape all user-provided content
                var escapedName = escapeHtml_{container_id}(name);
                var escapedDesc = escapeHtml_{container_id}(description);
                var escapedEmail = escapeHtml_{container_id}(email);
                
                tr.innerHTML = '<td><input type="checkbox" class="checkbox" disabled></td>' +
                    '<td>' + displayIndex + '</td>' +
                    '<td><div class="truncate" style="font-weight: 500;" title="' + escapedName + '">' + escapedName + '</div></td>' +
                    '<td><div class="truncate" style="color: #6b7280;" title="' + escapedDesc + '">' + escapedDesc + '</div></td>' +
                    '<td><div class="admin-email"><svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><path d="M19 21v-2a4 4 0 0 0-4-4H9a4 4 0 0 0-4 4v2"></path><circle cx="12" cy="7" r="4"></circle></svg><span class="truncate">' + escapedEmail + '</span></div></td>' +
                    '<td><span class="uid-text">' + uidShort + '</span></td>' +
                    '<td><div class="date-text"><svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><rect width="18" height="18" x="3" y="4" rx="2" ry="2"></rect><line x1="16" x2="16" y1="2" y2="6"></line><line x1="8" x2="8" y1="2" y2="6"></line><line x1="3" x2="21" y1="10" y2="10"></line></svg><span class="truncate">' + created + '</span></div></td>' +
                    '<td><span class="type-badge">' + fileType + '</span></td>' +
                    '<td><div style="display: flex; gap: 0.125rem;"><button class="btn btn-slate btn-disabled" title="Edit mock file"><svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><circle cx="12" cy="12" r="10"></circle><line x1="2" x2="22" y1="12" y2="12"></line><path d="M12 2a15.3 15.3 0 0 1 4 10 15.3 15.3 0 0 1-4 10 15.3 15.3 0 0 1-4-10 15.3 15.3 0 0 1 4-10z"></path></svg><span>Mock</span></button><button class="btn btn-gray btn-disabled" title="Edit private file"><svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><rect width="18" height="11" x="3" y="11" rx="2" ry="2"></rect><path d="M7 11V7a5 5 0 0 1 10 0v4"></path></svg><span>Private</span></button></div></td>' +
                    '<td><div style="display: flex; gap: 0.125rem;"><button class="btn btn-blue btn-disabled" title="View object details">Info</button><button class="btn btn-purple btn-disabled" title="Copy local file path">Path</button><button class="btn btn-red btn-disabled" title="Delete object"><svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><path d="M3 6h18"></path><path d="M19 6v14c0 1-1 2-2 2H7c-1 0-2-1-2-2V6"></path><path d="M8 6V4c0-1 1-2 2-2h4c1 0 2 1 2 2v2"></path><line x1="10" x2="10" y1="11" y2="17"></line><line x1="14" x2="14" y1="11" y2="17"></line></svg></button></div></td>';
                tbody.appendChild(tr);
            }}
        }}
        
        // Search functionality
        window['{container_id}_filteredObjects'] = window['{container_id}_objects'];
        
        function filterObjects_{container_id}() {{
            var searchTerm = document.getElementById('{container_id}-search').value.toLowerCase();
            var adminFilter = document.getElementById('{container_id}-filter').value.toLowerCase();
            var allObjects = window['{container_id}_objects'];
            
            var filtered = allObjects.filter(function(obj) {{
                var matchesSearch = !searchTerm || 
                    (obj.name && obj.name.toLowerCase().includes(searchTerm)) ||
                    (obj.description && obj.description.toLowerCase().includes(searchTerm)) ||
                    (obj.uid && obj.uid.toLowerCase().includes(searchTerm));
                
                var matchesAdmin = !adminFilter || 
                    (obj.private_url && obj.private_url.toLowerCase().includes(adminFilter));
                
                return matchesSearch && matchesAdmin;
            }});
            
            window['{container_id}_filteredObjects'] = filtered;
            window['{container_id}_totalObjects'] = filtered.length;
            window['{container_id}_currentPage'] = 1;
            
            updateDisplay_{container_id}();
        }}
        
        function clearSearch_{container_id}() {{
            document.getElementById('{container_id}-search').value = '';
            document.getElementById('{container_id}-filter').value = '';
            filterObjects_{container_id}();
        }}
        
        function updateDisplay_{container_id}() {{
            var currentPage = window['{container_id}_currentPage'];
            var itemsPerPage = window['{container_id}_itemsPerPage'];
            var totalObjects = window['{container_id}_totalObjects'];
            var totalPages = Math.max(1, Math.ceil(totalObjects / itemsPerPage));
            
            // Update page info
            document.getElementById('{container_id}-page-info').textContent = 'Page ' + currentPage + ' of ' + totalPages;
            
            // Update buttons
            document.getElementById('{container_id}-prev').disabled = currentPage === 1;
            document.getElementById('{container_id}-next').disabled = currentPage === totalPages;
            
            // Update table
            var tbody = document.getElementById('{container_id}-tbody');
            tbody.innerHTML = '';
            
            var start = (currentPage - 1) * itemsPerPage;
            var end = Math.min(start + itemsPerPage, totalObjects);
            var objects = window['{container_id}_filteredObjects'];
            
            for (var i = start; i < end; i++) {{
                // Reverse the order to show newest first
                var reverseIndex = window['{container_id}_totalObjects'] - 1 - i;
                var obj = objects[reverseIndex];
                if (!obj) continue;
                
                var name = obj.name || 'Unnamed Object';
                var uid = obj.uid || '';
                var uidShort = uid.substring(0, 8) + '...';
                var created = obj.created_at || 'Unknown';
                var description = (obj.description || "Object '" + name + "' with explicit mock...").substring(0, 40) + '...';
                
                // Extract email from private_url
                var email = 'unknown@example.com';
                if (obj.private_url && obj.private_url.startsWith('syft://')) {{
                    var parts = obj.private_url.split('/');
                    if (parts.length >= 3) {{
                        email = parts[2];
                    }}
                }}
                
                // Determine file type from metadata or default
                var fileType = '.txt';
                if (obj.metadata && obj.metadata.file_extension) {{
                    fileType = obj.metadata.file_extension;
                }}
                
                var tr = document.createElement('tr');
                tr.onclick = function(idx) {{ return function() {{ copyObjectCode_{container_id}(idx); }}; }}(reverseIndex);
                tr.style.cursor = 'pointer';
                
                // Escape all user-provided content
                var escapedName = escapeHtml_{container_id}(name);
                var escapedDesc = escapeHtml_{container_id}(description);
                var escapedEmail = escapeHtml_{container_id}(email);
                
                tr.innerHTML = '<td><input type="checkbox" class="checkbox" disabled></td>' +
                    '<td>' + reverseIndex + '</td>' +
                    '<td><div class="truncate" style="font-weight: 500;" title="' + escapedName + '">' + escapedName + '</div></td>' +
                    '<td><div class="truncate" style="color: #6b7280;" title="' + escapedDesc + '">' + escapedDesc + '</div></td>' +
                    '<td><div class="admin-email"><svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><path d="M19 21v-2a4 4 0 0 0-4-4H9a4 4 0 0 0-4 4v2"></path><circle cx="12" cy="7" r="4"></circle></svg><span class="truncate">' + escapedEmail + '</span></div></td>' +
                    '<td><span class="uid-text">' + uidShort + '</span></td>' +
                    '<td><div class="date-text"><svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><rect width="18" height="18" x="3" y="4" rx="2" ry="2"></rect><line x1="16" x2="16" y1="2" y2="6"></line><line x1="8" x2="8" y1="2" y2="6"></line><line x1="3" x2="21" y1="10" y2="10"></line></svg><span class="truncate">' + created + '</span></div></td>' +
                    '<td><span class="type-badge">' + fileType + '</span></td>' +
                    '<td><div style="display: flex; gap: 0.125rem;"><button class="btn btn-slate btn-disabled" title="Edit mock file"><svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><circle cx="12" cy="12" r="10"></circle><line x1="2" x2="22" y1="12" y2="12"></line><path d="M12 2a15.3 15.3 0 0 1 4 10 15.3 15.3 0 0 1-4 10 15.3 15.3 0 0 1-4-10 15.3 15.3 0 0 1 4-10z"></path></svg><span>Mock</span></button><button class="btn btn-gray btn-disabled" title="Edit private file"><svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><rect width="18" height="11" x="3" y="11" rx="2" ry="2"></rect><path d="M7 11V7a5 5 0 0 1 10 0v4"></path></svg><span>Private</span></button></div></td>' +
                    '<td><div style="display: flex; gap: 0.125rem;"><button class="btn btn-blue btn-disabled" title="View object details">Info</button><button class="btn btn-purple btn-disabled" title="Copy local file path">Path</button><button class="btn btn-red btn-disabled" title="Delete object"><svg class="icon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><path d="M3 6h18"></path><path d="M19 6v14c0 1-1 2-2 2H7c-1 0-2-1-2-2V6"></path><path d="M8 6V4c0-1 1-2 2-2h4c1 0 2 1 2 2v2"></path><line x1="10" x2="10" y1="11" y2="17"></line><line x1="14" x2="14" y1="11" y2="17"></line></svg></button></div></td>';
                tbody.appendChild(tr);
            }}
        }}
        
        // Override changePage to use filtered objects
        function changePage_{container_id}(direction) {{
            var currentPage = window['{container_id}_currentPage'];
            var itemsPerPage = window['{container_id}_itemsPerPage'];
            var totalObjects = window['{container_id}_totalObjects'];
            var totalPages = Math.max(1, Math.ceil(totalObjects / itemsPerPage));
            
            currentPage += direction;
            if (currentPage < 1) currentPage = 1;
            if (currentPage > totalPages) currentPage = totalPages;
            
            window['{container_id}_currentPage'] = currentPage;
            updateDisplay_{container_id}();
        }}
        
        // Add event listeners for live search
        document.getElementById('{container_id}-search').addEventListener('input', filterObjects_{container_id});
        document.getElementById('{container_id}-filter').addEventListener('input', filterObjects_{container_id});
        
        </script>
        """
        
        return html
    
    def _generate_fallback_widget_broken(self):
        """Generate a local HTML widget that matches the iframe UI when server is unavailable"""
        import uuid
        from datetime import datetime
        
        container_id = f"syft_widget_{uuid.uuid4().hex[:8]}"
        
        # Ensure objects are loaded
        self._ensure_loaded()
        
        html = f"""
        <style>
        #{container_id} {{
            background: #ffffff;
            border: 1px solid #e5e7eb;
            border-radius: 0.375rem;
            font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, sans-serif;
            max-height: 400px;
            overflow: hidden;
            display: flex;
            flex-direction: column;
        }}
        #{container_id} .widget-header {{
            background: #ffffff;
            border-bottom: 1px solid #e5e7eb;
            padding: 0.5rem;
        }}
        #{container_id} .search-controls {{
            display: flex;
            gap: 0.25rem;
            flex-wrap: wrap;
        }}
        #{container_id} .search-input {{
            flex: 1;
            min-width: 150px;
            padding: 0.25rem 0.5rem 0.25rem 1.75rem;
            border: 1px solid #d1d5db;
            border-radius: 0.25rem;
            font-size: 0.75rem;
            position: relative;
        }}
        #{container_id} .search-wrapper {{
            position: relative;
            flex: 1;
            min-width: 150px;
        }}
        #{container_id} .search-icon {{
            position: absolute;
            left: 0.5rem;
            top: 50%;
            transform: translateY(-50%);
            width: 0.75rem;
            height: 0.75rem;
            color: #9ca3af;
        }}
        #{container_id} .filter-input {{
            flex: 1;
            min-width: 150px;
            padding: 0.25rem 0.5rem 0.25rem 1.75rem;
            border: 1px solid #d1d5db;
            border-radius: 0.25rem;
            font-size: 0.75rem;
        }}
        #{container_id} .filter-wrapper {{
            position: relative;
            flex: 1;
            min-width: 150px;
        }}
        #{container_id} .filter-icon {{
            position: absolute;
            left: 0.5rem;
            top: 50%;
            transform: translateY(-50%);
            width: 0.75rem;
            height: 0.75rem;
            color: #9ca3af;
        }}
        #{container_id} button {{
            padding: 0.25rem 0.5rem;
            border-radius: 0.25rem;
            font-size: 0.75rem;
            border: none;
            cursor: pointer;
            transition: background-color 0.2s;
        }}
        #{container_id} .btn-primary {{
            background: #dbeafe;
            color: #1e40af;
        }}
        #{container_id} .btn-primary:hover {{
            background: #bfdbfe;
        }}
        #{container_id} .btn-secondary {{
            background: #f3f4f6;
            color: #374151;
        }}
        #{container_id} .btn-secondary:hover {{
            background: #e5e7eb;
        }}
        #{container_id} .btn-success {{
            background: #d1fae5;
            color: #065f46;
        }}
        #{container_id} .btn-success:hover {{
            background: #a7f3d0;
        }}
        #{container_id} .btn-gray {{
            background: #f3f4f6;
            color: #1f2937;
        }}
        #{container_id} .btn-gray:hover {{
            background: #e5e7eb;
        }}
        #{container_id} .table-container {{
            flex: 1;
            overflow: auto;
            border-bottom: 1px solid #e5e7eb;
        }}
        #{container_id} table {{
            width: 100%;
            border-collapse: collapse;
            font-size: 0.75rem;
            min-width: 1200px;
        }}
        #{container_id} thead {{
            background: #fafafa;
            border-bottom: 1px solid #e5e7eb;
            position: sticky;
            top: 0;
            z-index: 10;
        }}
        #{container_id} th {{
            text-align: left;
            padding: 0.375rem 0.25rem;
            font-weight: 500;
            color: #374151;
            cursor: pointer;
            user-select: none;
        }}
        #{container_id} th:hover {{
            background: #f3f4f6;
        }}
        #{container_id} tbody tr {{
            border-bottom: 1px solid #f3f4f6;
            transition: background-color 0.15s;
            cursor: pointer;
        }}
        #{container_id} tbody tr:hover {{
            background: #fafafa;
        }}
        #{container_id} tbody tr.selected {{
            background: #dbeafe;
        }}
        #{container_id} td {{
            padding: 0.375rem 0.25rem;
            color: #374151;
        }}
        #{container_id} .checkbox-cell {{
            width: 1.5rem;
            text-align: center;
        }}
        #{container_id} .index-cell {{
            width: 2rem;
            text-align: center;
            font-weight: 500;
        }}
        #{container_id} .name-cell {{
            font-weight: 500;
            color: #111827;
        }}
        #{container_id} .truncate {{
            overflow: hidden;
            text-overflow: ellipsis;
            white-space: nowrap;
            max-width: 200px;
        }}
        #{container_id} .admin-cell {{
            font-family: monospace;
            font-size: 0.75rem;
        }}
        #{container_id} .uid-cell {{
            font-family: monospace;
            font-size: 0.75rem;
            color: #6b7280;
        }}
        #{container_id} .date-cell {{
            color: #6b7280;
            font-size: 0.7rem;
        }}
        #{container_id} .type-badge {{
            display: inline-block;
            padding: 0.125rem 0.25rem;
            background: #f3f4f6;
            color: #1f2937;
            border-radius: 0.25rem;
            font-size: 0.75rem;
            font-weight: 500;
        }}
        #{container_id} .file-button {{
            padding: 0.125rem 0.375rem;
            margin-right: 0.125rem;
            border-radius: 0.25rem;
            font-size: 0.7rem;
            border: none;
            cursor: pointer;
            transition: background-color 0.2s;
        }}
        #{container_id} .mock-btn {{
            background: #e0e7ff;
            color: #4338ca;
        }}
        #{container_id} .mock-btn:hover {{
            background: #c7d2fe;
        }}
        #{container_id} .private-btn {{
            background: #f3f4f6;
            color: #374151;
        }}
        #{container_id} .private-btn:hover {{
            background: #e5e7eb;
        }}
        #{container_id} .action-btn {{
            padding: 0.125rem 0.375rem;
            margin-right: 0.125rem;
            border-radius: 0.25rem;
            font-size: 0.7rem;
            border: none;
            cursor: pointer;
            transition: background-color 0.2s;
        }}
        #{container_id} .info-btn {{
            background: #dbeafe;
            color: #1e40af;
        }}
        #{container_id} .info-btn:hover {{
            background: #bfdbfe;
        }}
        #{container_id} .path-btn {{
            background: #ede9fe;
            color: #7c3aed;
        }}
        #{container_id} .path-btn:hover {{
            background: #ddd6fe;
        }}
        #{container_id} .delete-btn {{
            background: #fee2e2;
            color: #dc2626;
        }}
        #{container_id} .delete-btn:hover {{
            background: #fecaca;
        }}
        #{container_id} .status-bar {{
            padding: 0.5rem;
            background: #fafafa;
            font-size: 0.75rem;
            color: #6b7280;
            display: none;
        }}
        #{container_id} .server-warning {{
            background: #fef3c7;
            color: #92400e;
            padding: 0.5rem;
            font-size: 0.75rem;
            display: flex;
            align-items: center;
            gap: 0.25rem;
        }}
        #{container_id} .modal {{
            display: none;
            position: fixed;
            z-index: 1000;
            left: 0;
            top: 0;
            width: 100%;
            height: 100%;
            background-color: rgba(0,0,0,0.4);
        }}
        #{container_id} .modal-content {{
            background-color: #fefefe;
            margin: 5% auto;
            padding: 20px;
            border: 1px solid #888;
            border-radius: 8px;
            width: 80%;
            max-width: 600px;
            max-height: 80vh;
            overflow-y: auto;
            position: relative;
        }}
        #{container_id} .modal-close {{
            color: #aaa;
            float: right;
            font-size: 28px;
            font-weight: bold;
            cursor: pointer;
            line-height: 20px;
        }}
        #{container_id} .modal-close:hover,
        #{container_id} .modal-close:focus {{
            color: #000;
        }}
        #{container_id} .modal-title {{
            font-size: 1.2rem;
            font-weight: bold;
            margin-bottom: 1rem;
            color: #1f2937;
        }}
        #{container_id} .modal-data {{
            background: #f3f4f6;
            padding: 1rem;
            border-radius: 4px;
            font-family: monospace;
            font-size: 0.875rem;
            white-space: pre-wrap;
            word-break: break-all;
            color: #374151;
        }}
        #{container_id} .modal-error {{
            background: #fee2e2;
            color: #dc2626;
            padding: 1rem;
            border-radius: 4px;
            font-size: 0.875rem;
        }}
        </style>
        
        <div id="{container_id}">
            <div class="widget-header">
                <div class="search-controls">
                    <div class="search-wrapper">
                        <svg class="search-icon" xmlns="http://www.w3.org/2000/svg" fill="none" stroke="currentColor" viewBox="0 0 24 24">
                            <circle cx="11" cy="11" r="8"></circle>
                            <path d="m21 21-4.3-4.3"></path>
                        </svg>
                        <input type="text" class="search-input" placeholder="Search objects..." id="{container_id}-search">
                    </div>
                    <div class="filter-wrapper">
                        <svg class="filter-icon" xmlns="http://www.w3.org/2000/svg" fill="none" stroke="currentColor" viewBox="0 0 24 24">
                            <polygon points="22 3 2 3 10 12.46 10 19 14 21 14 12.46 22 3"></polygon>
                        </svg>
                        <input type="text" class="filter-input" placeholder="Filter by Admin..." id="{container_id}-filter">
                    </div>
                    <button class="btn-primary" data-action="search">Search</button>
                    <button class="btn-secondary" data-action="clear">Clear</button>
                    <button class="btn-success" data-action="new">New</button>
                    <button class="btn-primary" data-action="selectAll">Select All</button>
                    <button class="btn-primary" data-action="generateCode">Generate Code</button>
                    <button class="btn-gray" data-action="openWindow">Open in Window</button>
                    <button class="btn-gray" title="Reinstall requires server connection" data-action="reinstall">🔄</button>
                </div>
            </div>
            <div class="table-container">
                <table>
                    <thead>
                        <tr>
                            <th class="checkbox-cell">
                                <input type="checkbox" onchange="toggleAllCheckboxes_{container_id}(this)">
                            </th>
                            <th class="index-cell">#</th>
                            <th>Name</th>
                            <th>Description</th>
                            <th>Admin</th>
                            <th>UID</th>
                            <th>Created</th>
                            <th>Type</th>
                            <th>Files</th>
                            <th>Actions</th>
                        </tr>
                    </thead>
                    <tbody>
        """
        
        # Add rows for each object
        for i, obj in enumerate(self._objects):
            # Use original index if this is a sliced collection, otherwise use current index
            display_index = self._original_indices[i] if self._original_indices else i
            email = self._get_object_email(obj)
            name = obj.name or "Unnamed Object"
            description = obj.description or f"Auto-generated object: {name}"
            uid = str(obj.uid)
            uid_short = uid[:8] + "..."
            
            # Format dates
            created_str = obj.created_at.strftime("%m/%d/%Y, %H:%M:%S") if hasattr(obj, 'created_at') and obj.created_at else ""
            
            # Determine type (simplified)
            obj_type = "—"
            if hasattr(obj, 'file_extension'):
                obj_type = obj.file_extension or "—"
            
            # Escape HTML in strings
            import html as html_module
            name_escaped = html_module.escape(name)
            desc_escaped = html_module.escape(description)
            email_escaped = html_module.escape(email)
            
            html += f"""
                        <tr data-index="{display_index}" data-name="{name_escaped.lower()}" data-email="{email_escaped.lower()}" 
                            data-desc="{desc_escaped.lower()}" onclick="toggleRowSelect_{container_id}(this, event)">
                            <td class="checkbox-cell" onclick="event.stopPropagation()">
                                <input type="checkbox" onchange="updateSelection_{container_id}()">
                            </td>
                            <td class="index-cell">{display_index}</td>
                            <td class="name-cell truncate" title="{name_escaped}">{name_escaped}</td>
                            <td class="truncate" title="{desc_escaped}">{desc_escaped}</td>
                            <td class="admin-cell truncate" title="{email_escaped}">
                                <span style="display: inline-flex; align-items: center; gap: 0.25rem;">
                                    <svg style="width: 0.75rem; height: 0.75rem;" xmlns="http://www.w3.org/2000/svg" fill="none" stroke="currentColor" viewBox="0 0 24 24">
                                        <path d="M19 21v-2a4 4 0 0 0-4-4H9a4 4 0 0 0-4 4v2"></path>
                                        <circle cx="12" cy="7" r="4"></circle>
                                    </svg>
                                    {email_escaped}
                                </span>
                            </td>
                            <td class="uid-cell" title="{uid}">{uid_short}</td>
                            <td class="date-cell">{created_str}</td>
                            <td><span class="type-badge">{obj_type}</span></td>
                            <td onclick="event.stopPropagation()">
                                <button class="file-button mock-btn" onclick="showAccessCode_{container_id}({display_index}, 'mock')">Mock</button>
                                <button class="file-button private-btn" onclick="showAccessCode_{container_id}({display_index}, 'private')">Private</button>
                            </td>
                            <td onclick="event.stopPropagation()">
                                <button class="action-btn info-btn" onclick="showObjectInfo_{container_id}({display_index})">Info</button>
                                <button class="action-btn path-btn" onclick="copyPath_{container_id}('{html_module.escape(str(obj.private_path))}')">Path</button>
                                <button class="action-btn delete-btn" onclick="confirmDelete_{container_id}({display_index})">
                                    <svg style="width: 0.75rem; height: 0.75rem; display: inline;" xmlns="http://www.w3.org/2000/svg" fill="none" stroke="currentColor" viewBox="0 0 24 24">
                                        <path d="M3 6h18"></path>
                                        <path d="M19 6v14c0 1-1 2-2 2H7c-1 0-2-1-2-2V6"></path>
                                        <path d="M8 6V4c0-1 1-2 2-2h4c1 0 2 1 2 2v2"></path>
                                    </svg>
                                </button>
                            </td>
                        </tr>
            """
        
        html += f"""
                    </tbody>
                </table>
            </div>
            <div class="status-bar" id="{container_id}-status">
                0 objects selected
            </div>
            
            <!-- Modal for displaying data -->
            <div id="{container_id}-modal" class="modal">
                <div class="modal-content">
                    <span class="modal-close" onclick="closeModal_{container_id}()">&times;</span>
                    <div class="modal-title" id="{container_id}-modal-title">Data Viewer</div>
                    <div id="{container_id}-modal-body"></div>
                </div>
            </div>
        </div>
        
        <script>
        // Immediately execute to set up the widget
        (function() {{
            console.log('Setting up widget {container_id}');
            
            // Make functions globally available for onclick handlers
            window.syftWidget_{container_id} = {{
                filterTable: function(searchTerm) {{
                    console.log('Filtering table with:', searchTerm);
            const rows = document.querySelectorAll('#{container_id} tbody tr');
            const term = searchTerm.toLowerCase();
            
            rows.forEach(row => {{
                const name = row.dataset.name || '';
                const email = row.dataset.email || '';
                const desc = row.dataset.desc || '';
                const visible = !term || name.includes(term) || email.includes(term) || desc.includes(term);
                row.style.display = visible ? '' : 'none';
            }});
            
            updateSelection_{container_id}();
        }}
        
        function filterByAdmin_{container_id}(adminTerm) {{
            const rows = document.querySelectorAll('#{container_id} tbody tr');
            const term = adminTerm.toLowerCase();
            
            rows.forEach(row => {{
                const email = row.dataset.email || '';
                const visible = !term || email.includes(term);
                row.style.display = visible ? '' : 'none';
            }});
            
            updateSelection_{container_id}();
        }}
        
        function searchFallback_{container_id}() {{
            const searchInput = document.querySelector('#{container_id} .search-input');
            const filterInput = document.querySelector('#{container_id} .filter-input');
            
            if (searchInput.value) {{
                filterFallbackTable_{container_id}(searchInput.value);
            }}
            if (filterInput.value) {{
                filterByAdmin_{container_id}(filterInput.value);
            }}
        }}
        
        function clearFallback_{container_id}() {{
            document.querySelector('#{container_id} .search-input').value = '';
            document.querySelector('#{container_id} .filter-input').value = '';
            filterFallbackTable_{container_id}('');
            
            // Clear all selections
            document.querySelectorAll('#{container_id} tbody input[type="checkbox"]').forEach(cb => {{
                cb.checked = false;
            }});
            document.querySelectorAll('#{container_id} tbody tr').forEach(row => {{
                row.classList.remove('selected');
            }});
            updateSelection_{container_id}();
        }}
        
        function selectAllFallback_{container_id}() {{
            const visibleRows = document.querySelectorAll('#{container_id} tbody tr:not([style*="display: none"])');
            const visibleCheckboxes = [];
            visibleRows.forEach(row => {{
                const cb = row.querySelector('input[type="checkbox"]');
                if (cb) visibleCheckboxes.push(cb);
            }});
            
            const allChecked = visibleCheckboxes.every(cb => cb.checked);
            
            visibleCheckboxes.forEach(cb => {{
                cb.checked = !allChecked;
                cb.closest('tr').classList.toggle('selected', !allChecked);
            }});
            
            updateSelection_{container_id}();
        }}
        
        function toggleAllCheckboxes_{container_id}(source) {{
            const checkboxes = document.querySelectorAll('#{container_id} tbody input[type="checkbox"]');
            checkboxes.forEach(cb => {{
                if (cb.closest('tr').style.display !== 'none') {{
                    cb.checked = source.checked;
                    cb.closest('tr').classList.toggle('selected', source.checked);
                }}
            }});
            updateSelection_{container_id}();
        }}
        
        function toggleRowSelect_{container_id}(row, event) {{
            if (event.target.tagName === 'TD' && !event.target.classList.contains('checkbox-cell')) {{
                const checkbox = row.querySelector('input[type="checkbox"]');
                if (checkbox) {{
                    checkbox.checked = !checkbox.checked;
                    row.classList.toggle('selected', checkbox.checked);
                    updateSelection_{container_id}();
                }}
            }}
        }}
        
        function updateSelection_{container_id}() {{
            const selected = document.querySelectorAll('#{container_id} tbody input[type="checkbox"]:checked').length;
            const total = document.querySelectorAll('#{container_id} tbody tr:not([style*="display: none"])').length;
            const statusBar = document.getElementById('{container_id}-status');
            
            if (selected > 0) {{
                statusBar.style.display = 'block';
                statusBar.textContent = `${{selected}} object(s) selected out of ${{total}} visible`;
            }} else {{
                statusBar.style.display = 'none';
            }}
        }}
        
        function showObjectInfo_{container_id}(index) {{
            const modal = document.getElementById('{container_id}-modal');
            const modalTitle = document.getElementById('{container_id}-modal-title');
            const modalBody = document.getElementById('{container_id}-modal-body');
            const row = document.querySelector(`#{container_id} tbody tr[data-index="${{index}}"]`);
            
            if (row) {{
                const name = row.querySelector('.name-cell').textContent || 'Unnamed';
                const uid = row.querySelector('.uid-cell').getAttribute('title') || 'Unknown';
                const created = row.querySelector('.date-cell').textContent || 'Unknown';
                const desc = row.querySelector('td:nth-child(4)').getAttribute('title') || 'No description';
                const email = row.querySelector('.admin-cell').textContent.trim() || 'Unknown';
                const type = row.querySelector('.type-badge').textContent || '—';
                
                modalTitle.textContent = `Object Information - ${{name}}`;
                
                const objects = {self._objects_data_json()};
                const obj = objects[index];
                
                let metadataHtml = '';
                if (obj && obj.metadata) {{
                    metadataHtml = '<strong>Metadata:</strong><br>';
                    for (const [key, value] of Object.entries(obj.metadata)) {{
                        metadataHtml += `${{key}}: ${{value}}<br>`;
                    }}
                }}
                
                modalBody.innerHTML = `
                    <div style="font-size: 0.9rem; line-height: 1.6;">
                        <strong>Name:</strong> ${{name}}<br>
                        <strong>UID:</strong> <code style="background: #f3f4f6; padding: 2px 4px; border-radius: 3px;">${{uid}}</code><br>
                        <strong>Admin Email:</strong> ${{email}}<br>
                        <strong>Created:</strong> ${{created}}<br>
                        <strong>File Type:</strong> ${{type}}<br>
                        <strong>Description:</strong> ${{desc}}<br><br>
                        ${{metadataHtml}}
                        <br>
                        <strong>Private URL:</strong> <code style="background: #f3f4f6; padding: 2px 4px; border-radius: 3px; word-break: break-all;">${{obj ? obj.private_url : 'N/A'}}</code><br>
                        <strong>Mock URL:</strong> <code style="background: #f3f4f6; padding: 2px 4px; border-radius: 3px; word-break: break-all;">${{obj ? obj.mock_url : 'N/A'}}</code>
                    </div>
                `;
                
                modal.style.display = 'block';
            }}
        }}
        
        function copyPath_{container_id}(path) {{
            // Try to copy to clipboard
            if (navigator.clipboard) {{
                navigator.clipboard.writeText(path).then(() => {{
                    alert('Path copied to clipboard: ' + path);
                }}).catch(() => {{
                    prompt('Copy this path:', path);
                }});
            }} else {{
                prompt('Copy this path:', path);
            }}
        }}
        
        function showAccessCode_{container_id}(index, type) {{
            const modal = document.getElementById('{container_id}-modal');
            const modalTitle = document.getElementById('{container_id}-modal-title');
            const modalBody = document.getElementById('{container_id}-modal-body');
            const row = document.querySelector(`#{container_id} tbody tr[data-index="${{index}}"]`);
            
            if (row) {{
                const name = row.querySelector('.name-cell').textContent || 'Unnamed';
                const uid = row.querySelector('.uid-cell').getAttribute('title') || 'Unknown';
                
                modalTitle.textContent = `${{type.charAt(0).toUpperCase() + type.slice(1)}} Data - ${{name}}`;
                
                // Try to get the data from the object
                const objects = {self._objects_data_json()};
                const obj = objects[index];
                
                if (obj && obj[type + '_data']) {{
                    // Create text node to safely display content
                    const dataDiv = document.createElement('div');
                    dataDiv.className = 'modal-data';
                    dataDiv.textContent = obj[type + '_data'];
                    modalBody.innerHTML = '';
                    modalBody.appendChild(dataDiv);
                }} else {{
                    modalBody.innerHTML = `
                        <div class="modal-error">
                            <strong>Data not available</strong><br><br>
                            The ${{type}} data for this object is not cached locally. 
                            To access this data, please ensure the syft-objects server is running.<br><br>
                            <strong>Object Details:</strong><br>
                            Name: ${{name}}<br>
                            UID: ${{uid}}<br>
                            Type: ${{type}}
                        </div>
                    `;
                }}
                
                modal.style.display = 'block';
            }}
        }}
        
        function closeModal_{container_id}() {{
            const modal = document.getElementById('{container_id}-modal');
            modal.style.display = 'none';
        }}
        
        // Close modal when clicking outside of it
        window.addEventListener('click', function(event) {{
            const modal = document.getElementById('{container_id}-modal');
            if (event.target === modal) {{
                modal.style.display = 'none';
            }}
        }});
        
        function generateCode_{container_id}() {{
            const selectedCheckboxes = document.querySelectorAll(`#{container_id} tbody input[type="checkbox"]:checked`);
            const selectedIndices = [];
            
            selectedCheckboxes.forEach(cb => {{
                const row = cb.closest('tr');
                if (row) {{
                    const index = parseInt(row.dataset.index);
                    selectedIndices.push(index);
                }}
            }});
            
            if (selectedIndices.length === 0) {{
                alert('Please select at least one object to generate code.');
                return;
            }}
            
            // Generate Python code
            let code = '';
            if (selectedIndices.length === 1) {{
                code = `# Access single object\\nobj = so.objects[${{selectedIndices[0]}}]\\n\\n# Access data\\nobj.private  # Private data\\nobj.mock     # Mock data`;
            }} else {{
                const indicesStr = '[' + selectedIndices.join(', ') + ']';
                code = `# Access multiple objects\\nselected_objects = so.objects.get_by_indices(${{indicesStr}})\\n\\n# Access data for each object\\nfor obj in selected_objects:\\n    print(f"Object: {{obj.name}}")\\n    # obj.private  # Private data\\n    # obj.mock     # Mock data`;
            }}
            
            // Show code in modal
            const modal = document.getElementById('{container_id}-modal');
            const modalTitle = document.getElementById('{container_id}-modal-title');
            const modalBody = document.getElementById('{container_id}-modal-body');
            
            modalTitle.textContent = 'Generated Code';
            modalBody.innerHTML = `
                <div style="margin-bottom: 10px;">
                    <button onclick="copyGeneratedCode_{container_id}()" style="padding: 5px 10px; background: #dbeafe; color: #1e40af; border: none; border-radius: 4px; cursor: pointer;">
                        Copy to Clipboard
                    </button>
                </div>
                <pre style="background: #f3f4f6; padding: 1rem; border-radius: 4px; overflow-x: auto;">
<code id="{container_id}-generated-code">${{code}}</code>
                </pre>
            `;
            
            modal.style.display = 'block';
        }}
        
        function copyGeneratedCode_{container_id}() {{
            const codeElement = document.getElementById('{container_id}-generated-code');
            const code = codeElement.textContent;
            
            if (navigator.clipboard) {{
                navigator.clipboard.writeText(code).then(() => {{
                    alert('Code copied to clipboard!');
                }}).catch(() => {{
                    fallbackCopy(code);
                }});
            }} else {{
                fallbackCopy(code);
            }}
            
            function fallbackCopy(text) {{
                const textarea = document.createElement('textarea');
                textarea.value = text;
                textarea.style.position = 'fixed';
                textarea.style.opacity = '0';
                document.body.appendChild(textarea);
                textarea.select();
                try {{
                    document.execCommand('copy');
                    alert('Code copied to clipboard!');
                }} catch (err) {{
                    prompt('Copy this code:', text);
                }}
                document.body.removeChild(textarea);
            }}
        }}
        
        function confirmDelete_{container_id}(index) {{
            const row = document.querySelector(`#{container_id} tbody tr[data-index="${{index}}"]`);
            if (!row) return;
            
            const name = row.querySelector('.name-cell').textContent || 'Unnamed Object';
            const uid = row.querySelector('.uid-cell').getAttribute('title') || 'Unknown';
            
            const modal = document.getElementById('{container_id}-modal');
            const modalTitle = document.getElementById('{container_id}-modal-title');
            const modalBody = document.getElementById('{container_id}-modal-body');
            
            modalTitle.textContent = 'Confirm Deletion';
            modalBody.innerHTML = `
                <div style="font-size: 0.9rem;">
                    <p style="margin-bottom: 1rem;">Are you sure you want to delete this object?</p>
                    <div style="background: #fee2e2; padding: 1rem; border-radius: 4px; margin-bottom: 1rem;">
                        <strong>Object to delete:</strong><br>
                        Name: ${{name}}<br>
                        UID: ${{uid}}
                    </div>
                    <p style="color: #dc2626; font-weight: bold;">This action cannot be undone!</p>
                    <div style="margin-top: 1.5rem; display: flex; gap: 0.5rem; justify-content: flex-end;">
                        <button onclick="closeModal_{container_id}()" style="padding: 0.5rem 1rem; background: #e5e7eb; color: #374151; border: none; border-radius: 4px; cursor: pointer;">
                            Cancel
                        </button>
                        <button onclick="deleteObject_{container_id}(${{index}})" style="padding: 0.5rem 1rem; background: #dc2626; color: white; border: none; border-radius: 4px; cursor: pointer;">
                            Delete
                        </button>
                    </div>
                </div>
            `;
            
            modal.style.display = 'block';
        }}
        
        function deleteObject_{container_id}(index) {{
            // In offline mode, we can only remove from display
            const row = document.querySelector(`#{container_id} tbody tr[data-index="${{index}}"]`);
            if (row) {{
                row.style.transition = 'opacity 0.3s';
                row.style.opacity = '0';
                setTimeout(() => {{
                    row.remove();
                    updateSelection_{container_id}();
                    // Re-index remaining rows
                    const rows = document.querySelectorAll('#{container_id} tbody tr');
                    rows.forEach((r, i) => {{
                        r.querySelector('.index-cell').textContent = i;
                    }});
                }}, 300);
            }}
            
            closeModal_{container_id}();
            
            // Show notification
            alert('Object marked for deletion. Note: Actual deletion requires server connection.');
        }}
        </script>
        """
        
        return html

    def widget(self, width="100%", height="400px", url=None):
        """Display the syft-objects widget in an iframe"""
        
        self._ensure_server_ready()
        if url is None:
            url = get_syft_objects_url("widget")
        
        return f"""
        <iframe 
            src="{url}" 
            width="{width}" 
            height="{height}"
            frameborder="0"
            style="border: none;"
            title="SyftObjects Widget">
        </iframe>
        """

    def _generate_interactive_table_html(self, title, count, search_indicator, container_id):
        """Generate the interactive HTML table"""
        html = f"""
        <style>
        .syft-objects-container {{
            max-height: 500px;
            border: 1px solid #dee2e6;
            border-radius: 6px;
            margin: 10px 0;
            font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, sans-serif;
        }}
        .syft-objects-header {{
            background-color: #f8f9fa;
            padding: 10px 15px;
            border-bottom: 1px solid #dee2e6;
            margin: 0;
        }}
        .syft-objects-controls {{
            padding: 10px 15px;
            background-color: #fff;
            border-bottom: 1px solid #dee2e6;
            display: flex;
            gap: 10px;
            align-items: center;
        }}
        .syft-objects-search-box {{
            flex: 1;
            padding: 6px 10px;
            border: 1px solid #ced4da;
            border-radius: 4px;
            font-size: 12px;
        }}
        .syft-objects-btn {{
            padding: 6px 12px;
            background-color: #007bff;
            color: white;
            border: none;
            border-radius: 4px;
            cursor: pointer;
            font-size: 11px;
            text-decoration: none;
        }}
        .syft-objects-btn:hover {{
            background-color: #0056b3;
        }}
        .syft-objects-btn-secondary {{
            background-color: #6c757d;
        }}
        .syft-objects-btn-secondary:hover {{
            background-color: #545b62;
        }}
        .syft-objects-table-container {{
            max-height: 320px;
            overflow-y: auto;
            overflow-x: auto;
        }}
        .syft-objects-table {{
            border-collapse: collapse;
            width: 100%;
            font-size: 11px;
            margin: 0;
            min-width: 1400px;
        }}
        .syft-objects-table th {{
            background-color: #f8f9fa;
            border-bottom: 2px solid #dee2e6;
            padding: 6px 8px;
            text-align: left;
            font-weight: 600;
            color: #495057;
            position: sticky;
            top: 0;
            z-index: 10;
        }}
        .syft-objects-table td {{
            border-bottom: 1px solid #f1f3f4;
            padding: 4px 8px;
            vertical-align: top;
        }}
        .syft-objects-table tr:hover {{
            background-color: #f8f9fa;
        }}
        .syft-objects-table tr.syft-objects-selected {{
            background-color: #e3f2fd;
        }}
        .syft-objects-email {{
            color: #0066cc;
            font-weight: 500;
            font-size: 10px;
            min-width: 120px;
        }}
        .syft-objects-name {{
            color: #28a745;
            font-weight: 500;
            min-width: 150px;
        }}
        .syft-objects-url {{
            font-family: 'Monaco', 'Menlo', 'Ubuntu Mono', monospace;
            font-size: 9px;
            color: #6c757d;
            min-width: 200px;
            word-break: break-all;
        }}
        .syft-objects-metadata {{
            font-family: 'Monaco', 'Menlo', 'Ubuntu Mono', monospace;
            font-size: 9px;
            color: #8b5cf6;
            min-width: 180px;
            max-width: 320px;
            word-break: break-all;
            white-space: pre-wrap;
        }}
        .syft-objects-desc {{
            font-size: 10px;
            color: #374151;
            min-width: 180px;
            max-width: 320px;
            word-break: break-word;
            white-space: pre-wrap;
        }}
        .syft-objects-date {{
            font-size: 10px;
            color: #64748b;
            min-width: 120px;
            max-width: 160px;
            word-break: break-word;
        }}
        .syft-objects-index {{
            text-align: center;
            font-weight: 600;
            color: #495057;
            background-color: #f8f9fa;
            width: 40px;
            min-width: 40px;
        }}
        .syft-objects-checkbox {{
            width: 40px;
            min-width: 40px;
            text-align: center;
        }}
        .syft-objects-output {{
            padding: 10px 15px;
            background-color: #f8f9fa;
            border-top: 1px solid #dee2e6;
            font-family: 'Monaco', 'Menlo', 'Ubuntu Mono', monospace;
            font-size: 10px;
            color: #495057;
            white-space: pre-wrap;
            overflow-x: auto;
        }}
        .syft-objects-status {{
            padding: 5px 15px;
            background-color: #e9ecef;
            font-size: 10px;
            color: #6c757d;
        }}
        </style>
        <div class="syft-objects-container" id="{container_id}">
            <div class="syft-objects-header">
                <strong>🔐 {title} ({count} total)</strong>
                {search_indicator}
            </div>
            <div class="syft-objects-controls">
                <input type="text" class="syft-objects-search-box" placeholder="🔍 Search objects..." 
                       onkeyup="filterSyftObjects('{container_id}')">
                <button class="syft-objects-btn" onclick="selectAllSyftObjects('{container_id}')">Select All</button>
                <button class="syft-objects-btn syft-objects-btn-secondary" onclick="clearAllSyftObjects('{container_id}')">Clear</button>
                <button class="syft-objects-btn" onclick="generateSyftObjectsCode('{container_id}')">Generate Code</button>
                <button class="syft-objects-btn" onclick="createNewSyftObject('{container_id}')">New</button>
            </div>
            <div class="syft-objects-table-container">
                <table class="syft-objects-table">
                    <thead>
                        <tr>
                            <th style="width: 40px; min-width: 40px;">☑</th>
                            <th style="width: 40px; min-width: 40px;">#</th>
                            <th style="min-width: 120px;">Email</th>
                            <th style="min-width: 150px;">Object Name</th>
                            <th style="min-width: 200px;">Private URL</th>
                            <th style="min-width: 200px;">Mock URL</th>
                            <th style="min-width: 120px;">Created</th>
                            <th style="min-width: 120px;">Updated</th>
                            <th style="min-width: 180px;">Description</th>
                            <th style="min-width: 180px;">Metadata</th>
                        </tr>
                    </thead>
                    <tbody>
        """

        for i, syft_obj in enumerate(self._objects):
            email = self._get_object_email(syft_obj)
            name = syft_obj.get_name() if hasattr(syft_obj, 'get_name') else (syft_obj.name if hasattr(syft_obj, 'name') else "Unnamed Object")
            # Compact metadata string (excluding system keys)
            system_keys = {"_file_operations"}
            metadata = syft_obj.get_metadata() if hasattr(syft_obj, 'get_metadata') else (syft_obj.metadata if hasattr(syft_obj, 'metadata') else {})
            meta_items = [f"{k}={v}" for k, v in metadata.items() if k not in system_keys]
            meta_str = ", ".join(meta_items) if meta_items else ""
            created_at = syft_obj.get_created_at() if hasattr(syft_obj, 'get_created_at') else (syft_obj.created_at if hasattr(syft_obj, 'created_at') else None)
            created_str = created_at.strftime("%Y-%m-%d %H:%M UTC") if created_at else ""
            updated_at = syft_obj.get_updated_at() if hasattr(syft_obj, 'get_updated_at') else (syft_obj.updated_at if hasattr(syft_obj, 'updated_at') else None)
            updated_str = updated_at.strftime("%Y-%m-%d %H:%M UTC") if updated_at else ""
            desc_str = syft_obj.get_description() if hasattr(syft_obj, 'get_description') else (syft_obj.description if hasattr(syft_obj, 'description') else "")
            html += f"""
            <tr data-email="{email.lower()}" data-name="{name.lower()}" data-index="{i}" data-meta="{meta_str.lower()}" data-desc="{desc_str.lower()}" data-created="{created_str.lower()}" data-updated="{updated_str.lower()}">
                <td class="syft-objects-checkbox">
                    <input type="checkbox" onchange="updateSyftObjectsSelection('{container_id}')">
                </td>
                <td class="syft-objects-index">{i}</td>
                <td class="syft-objects-email">{email}</td>
                <td class="syft-objects-name">{name}</td>
                <td class="syft-objects-url">{syft_obj.get_urls()['private'] if hasattr(syft_obj, 'get_urls') else (syft_obj.private_url if hasattr(syft_obj, 'private_url') else '')}</td>
                <td class="syft-objects-url">{syft_obj.get_urls()['mock'] if hasattr(syft_obj, 'get_urls') else (syft_obj.mock_url if hasattr(syft_obj, 'mock_url') else '')}</td>
                <td class="syft-objects-date">{created_str}</td>
                <td class="syft-objects-date">{updated_str}</td>
                <td class="syft-objects-desc">{desc_str}</td>
                <td class="syft-objects-metadata">{meta_str}</td>
            </tr>
            """

        html += f"""
                    </tbody>
                </table>
            </div>
            <div class="syft-objects-status" id="{container_id}-status">
                0 objects selected • Use checkboxes to select objects
            </div>
            <div class="syft-objects-output" id="{container_id}-output" style="display: none;">
                # Copy this code to your notebook:
            </div>
        </div>
        
        <script>
        function filterSyftObjects(containerId) {{
            const searchBox = document.querySelector(`#${{containerId}} .syft-objects-search-box`);
            const table = document.querySelector(`#${{containerId}} .syft-objects-table tbody`);
            const rows = table.querySelectorAll('tr');
            const searchTerm = searchBox.value.toLowerCase();
            
            let visibleCount = 0;
            rows.forEach(row => {{
                const email = row.dataset.email || '';
                const name = row.dataset.name || '';
                const meta = row.dataset.meta || '';
                const desc = row.dataset.desc || '';
                const created = row.dataset.created || '';
                const updated = row.dataset.updated || '';
                const isVisible = email.includes(searchTerm) || name.includes(searchTerm) || meta.includes(searchTerm) || desc.includes(searchTerm) || created.includes(searchTerm) || updated.includes(searchTerm);
                row.style.display = isVisible ? '' : 'none';
                if (isVisible) visibleCount++;
            }});
            
            updateSyftObjectsSelection(containerId);
        }}
        
        function selectAllSyftObjects(containerId) {{
            const table = document.querySelector(`#${{containerId}} .syft-objects-table tbody`);
            const checkboxes = table.querySelectorAll('input[type="checkbox"]');
            const visibleCheckboxes = Array.from(checkboxes).filter(cb => 
                cb.closest('tr').style.display !== 'none'
            );
            
            const allChecked = visibleCheckboxes.every(cb => cb.checked);
            visibleCheckboxes.forEach(cb => cb.checked = !allChecked);
            
            updateSyftObjectsSelection(containerId);
        }}
        
        function clearAllSyftObjects(containerId) {{
            const table = document.querySelector(`#${{containerId}} .syft-objects-table tbody`);
            const checkboxes = table.querySelectorAll('input[type="checkbox"]');
            checkboxes.forEach(cb => cb.checked = false);
            updateSyftObjectsSelection(containerId);
        }}
        
        function updateSyftObjectsSelection(containerId) {{
            const table = document.querySelector(`#${{containerId}} .syft-objects-table tbody`);
            const rows = table.querySelectorAll('tr');
            const status = document.querySelector(`#${{containerId}}-status`);
            
            let selectedCount = 0;
            rows.forEach(row => {{
                const checkbox = row.querySelector('input[type="checkbox"]');
                if (checkbox && checkbox.checked) {{
                    row.classList.add('syft-objects-selected');
                    selectedCount++;
                }} else {{
                    row.classList.remove('syft-objects-selected');
                }}
            }});
            
            const visibleRows = Array.from(rows).filter(row => row.style.display !== 'none');
            status.textContent = `${{selectedCount}} object(s) selected • ${{visibleRows.length}} visible`;
        }}
        
        function generateSyftObjectsCode(containerId) {{
            const table = document.querySelector(`#${{containerId}} .syft-objects-table tbody`);
            const rows = table.querySelectorAll('tr');
            const output = document.querySelector(`#${{containerId}}-output`);
            
            const selectedIndices = [];
            rows.forEach(row => {{
                const checkbox = row.querySelector('input[type="checkbox"]');
                if (checkbox && checkbox.checked) {{
                    selectedIndices.push(row.dataset.index);
                }}
            }});
            
            if (selectedIndices.length === 0) {{
                output.style.display = 'none';
                return;
            }}
            
            let code;
            if (selectedIndices.length === 1) {{
                code = `# Selected object:
obj = so.objects[${{selectedIndices[0]}}]`;
            }} else {{
                const indicesStr = selectedIndices.join(', ');
                code = `# Selected objects:
objects = [so.objects[i] for i in [${{indicesStr}}]]`;
            }}
            
            // Copy to clipboard
            navigator.clipboard.writeText(code).then(() => {{
                // Update button text to show success
                const button = document.querySelector(`#${{containerId}} button[onclick="generateSyftObjectsCode('${{containerId}}')"]`);
                const originalText = button.textContent;
                button.textContent = '✅ Copied!';
                button.style.backgroundColor = '#28a745';
                
                // Reset button after 2 seconds
                setTimeout(() => {{
                    button.textContent = originalText;
                    button.style.backgroundColor = '#007bff';
                }}, 2000);
            }}).catch(err => {{
                console.warn('Could not copy to clipboard:', err);
                // Fallback: still show the code for manual copying
            }});
            
            output.textContent = code;
            output.style.display = 'block';
        }}
        
        function createNewSyftObject(containerId) {{
            // Show confirmation message and provide template code
            const output = document.querySelector(`#${{containerId}}-output`);
            const code = `# Create a new SyftObject:
import syft as sy

# Example: Create a new object with your data
new_object = sy.SyftObject(
    name="My New Object",
    description="Description of my object",
    # Add your data and configuration here
)

# Upload to your datasite
# client.upload(new_object)`;
            
            // Copy to clipboard
            navigator.clipboard.writeText(code).then(() => {{
                // Update button text to show success
                const button = document.querySelector(`#${{containerId}} button[onclick="createNewSyftObject('${{containerId}}')"]`);
                const originalText = button.textContent;
                button.textContent = '✅ Template Copied!';
                button.style.backgroundColor = '#28a745';
                
                // Reset button after 2 seconds
                setTimeout(() => {{
                    button.textContent = originalText;
                    button.style.backgroundColor = '#007bff';
                }}, 2000);
            }}).catch(err => {{
                console.warn('Could not copy to clipboard:', err);
                // Fallback: still show the code for manual copying
            }});
            
            output.textContent = code;
            output.style.display = 'block';
            
            // Update status to show what happened
            const status = document.querySelector(`#${{containerId}}-status`);
            status.textContent = 'New object template generated • Copy the code above to create a new SyftObject';
        }}
        </script>
        """

        return html

    def delete_object(self, uid_or_index, user_email: str = None):
        """
        Delete an object by UID or index with permission checking.
        
        Args:
            uid_or_index: Either the UID string or integer index of the object to delete
            user_email: Email of the user attempting deletion. If None, will try to get from SyftBox client.
        
        Returns:
            bool: True if deletion was successful
            
        Raises:
            PermissionError: If user doesn't have permission to delete the object
            KeyError: If object with given UID is not found
            IndexError: If index is out of bounds
        """
        self._ensure_loaded()
        
        # Find the object to delete
        target_obj = None
        if isinstance(uid_or_index, str):
            # Delete by UID
            for obj in self._objects:
                if str(obj.uid) == uid_or_index:
                    target_obj = obj
                    break
            if not target_obj:
                raise KeyError(f"Object with UID '{uid_or_index}' not found")
        elif isinstance(uid_or_index, int):
            # Delete by index
            if not 0 <= uid_or_index < len(self._objects):
                raise IndexError(f"Index {uid_or_index} out of bounds (0-{len(self._objects)-1})")
            target_obj = self._objects[uid_or_index]
        else:
            raise ValueError("uid_or_index must be a string UID or integer index")
        
        # Perform the deletion with permission checking
        success = target_obj.delete_obj(user_email)
        
        # If successful, refresh the collection
        if success:
            self.refresh()
        
        return success 